import fs, { WriteStream } from "fs";
import https from "https";
import uuid from "uuid/v4";

interface ITranslateRequest { Text: string; originalKey: string; }

interface ITranslateResponse {
  translations: {
    text: string;
    to: string;
  }[];
}

interface ITranslation {
  de: string;
  en: string;
  fr: string;
  [key: string]: string;
}

interface ITranslationMap {
  [key: string]: ITranslation;
}

export interface ITranslationParams {
  key: string,
  inputDirectory: string,
  langs: string[],
  outputDir: string
}

export class ESOTranslate {
  private readonly SUPPORTED_LANGS: string[] = 'fr,de,jp,ru,it'.split(',');

  private static generateRequests(contentStr: string) {
    const json = JSON.parse(contentStr);
    return Object.keys(json).map<ITranslateRequest>((key) => {
      return {
        Text: json[key],
        originalKey: key,
      };
    }).reduce<ITranslateRequest[][]>(
      (accum: ITranslateRequest[][], current: ITranslateRequest) => {
        if (accum.length === 0) {
          accum[0] = [];
        }

        // Per docs, can only have up to 100 items in a request.
        // https://docs.microsoft.com/en-us/azure/cognitive-services/translator/reference/v3-0-translate#request-body
        // TODO check request size <= 5000 characters.
        if (accum[accum.length - 1].length >= 50) {
          accum[accum.length] = [];
        }

        accum[accum.length - 1].push(current);

        return accum;
      }, []);
  }

  private static readFile(filePath: string): Promise<string> {
    return new Promise((resolve, reject) => fs.readFile(filePath, "utf8", (error, data) => {
      if (error) {
        return reject(error);
      }

      resolve(data);
    }));
  }

  private readonly host = "api.cognitive.microsofttranslator.com";
  private readonly path = "/translate?api-version=3.0";
  private readonly requestParams: ITranslationParams;

  constructor(request: ITranslationParams) {
    if (!request.langs) {
      request.langs = this.SUPPORTED_LANGS;
    }
    this.requestParams = request;
  }

  public async translate() {
    const json = await ESOTranslate.readFile(`${this.requestParams.inputDirectory}/en.json`);
    const enStream = fs.createWriteStream(`${this.requestParams.outputDir}/en.lua`);
    console.log(`Writing to ${this.requestParams.outputDir}/en.lua`);
    const parsed: {[key: string]: string} = JSON.parse(json);
    Object.keys(parsed).forEach((key) => {
      enStream.write(`ZO_CreateStringId("${key}", "${parsed[key]}")\n`,
        (error) => {
          if (error) {
            throw error;
          } 
        });
    });

    const batchRequests: ITranslateRequest[][] = ESOTranslate.generateRequests(json);
    const batchTranslations: ITranslateResponse[][] =
      await Promise.all(batchRequests.map((request) => this.requestTranslations(request)))
        .then((response) => response.map((translation: any) => JSON.parse(translation)));

    return this.writeTranslations(await this.collate(batchRequests, batchTranslations));
  }

  private async collate(batchRequests: ITranslateRequest[][], batchTranslations: ITranslateResponse[][]) {
    const output: any = {};
    const overrides: any = await this.getOverrides();
    for (let batch = 0; batch < batchRequests.length; batch++) {
      const requests = batchRequests[batch];
      const translations = batchTranslations[batch];
      for (let i = 0; i < requests.length; i++) {
        const request = requests[i];
        const translation = translations[i];
        const key = request.originalKey;
        output[key] = { en: request.Text }
        translation.translations.forEach((localeTranslation) => {
          // Revert back to ESO abbreviations
          const lang = localeTranslation.to === "ja" ? "jp" : localeTranslation.to;
          if (overrides[lang] && overrides[lang][key]) {
            output[key][lang] = overrides[lang][key];
          } else {
            output[key][lang] = localeTranslation.text.replace(/"/g, "'");;
          }
        });
      }
    }

    return output;
  }

  private async getOverrides() {
    const promises: any = [];
    const overrides: any = {};
    this.requestParams.langs.forEach(async lang => {
      promises.push(new Promise<void>(async (resolve) => {
        try {
          const data = await ESOTranslate.readFile(`${this.requestParams.inputDirectory}/${lang}.json`);
          overrides[lang] = JSON.parse(data);
        } catch (error) {
          //
        }
        resolve();
      }))
    });
    return Promise.all(promises)
      .then(() => {
        return overrides;
      });
  }

  private async writeTranslations(translations: ITranslationMap) {
    await this.createOutputDirectory();
    const writeStreams: { [key: string]: WriteStream } = {};
    this.requestParams.langs.forEach((lang) => {
      console.log(`Writing to ${this.requestParams.outputDir}/${lang}.lua`);
      writeStreams[lang] = fs.createWriteStream(`${this.requestParams.outputDir}/${lang}.lua`);
    });
    await Promise.all(Object.keys(translations).map(async (key) => {
      await Promise.all(this.requestParams.langs.map(async (lang) => {
        await new Promise<void>((resolve, reject) =>
          writeStreams[lang].write(`ZO_CreateStringId("${key}", "${translations[key][lang]}")\n`,
            (error) => {
              if (error) {
                reject(error);
              } else {
                resolve();
              }
            }));
      }));
    }));
  }

  private async createOutputDirectory() {
    await new Promise<void>((resolve, reject) => {
      fs.exists(this.requestParams.outputDir, (exists) => {
        if (exists) {
          return resolve();
        }

        fs.mkdir(this.requestParams.outputDir, (error) => {
          if (error) {
            return reject(error);
          }

          resolve();
        });
      });
    });
  }

  private requestTranslations(request: ITranslateRequest[]) {
    return new Promise((resolve, reject) => {
      let params = '&from=en';
      this.requestParams.langs.forEach(lang => {
        if (lang !== "en") {
          if (lang === "jp") {
            // Azure uses ja for Japanese
            lang = "ja";
          }
        }

        params = params.concat(`&to=${lang}`);
      });
      const requestParams = {
        headers: {
          "Content-Type": "application/json",
          "Ocp-Apim-Subscription-Key": this.requestParams.key,
          "X-ClientTraceId": uuid(),
        },
        hostname: this.host,
        method: "POST",
        path: this.path + params,
      };

      const req = https.request(requestParams, (response) => {
        let body = "";
        response.on("data", (d) => {
          body += d;
        });

        response.on("end", () => {
          resolve(body);
        });

        response.on("error", (e) => {
          reject(e);
        });
      });

      req.write(JSON.stringify(request));
      req.end();
    });
  }
}
